//
//  CFRMandelbrotRender.m
//  Mandelbrot
//
//  Created by Renaud Pradenc on 03/02/09.
//  Copyright 2009 Céroce. All rights reserved.
//

#import "CFRMandelbrotRender.h"


@implementation CFRMandelbrotRender

#pragma mark Gestion de la classe

- (id) init
{
	if(self = [super init])
	{
		largeurBitmap = 400;
		hauteurBitmap = 300;

		largeur = 4.0;
		centre.reel = -0.7;
		centre.imag = 0.0;
		
		maxIterations = 18;
		besoinMAJ = NO;
	}
	
	return self;
}



#pragma mark Accesseurs
- (void) setLargeurBitmap:(NSUInteger)largeurPixels
{
	largeurBitmap = largeurPixels;	
	
	[self willChangeValueForKey:@"besoinMAJ"];
	besoinMAJ = YES;
	[self didChangeValueForKey:@"besoinMAJ"];
	
}

- (void) setHauteurBitmap:(NSUInteger)hauteurPixels
{
	hauteurBitmap = hauteurPixels;	
	
	[self willChangeValueForKey:@"besoinMAJ"];
	besoinMAJ = YES;
	[self didChangeValueForKey:@"besoinMAJ"];
}

- (void) setMaxIterations:(int)iterations
{
	if(iterations > 255)
		NSLog(@"-[CFRMandelbrotRender setMaxIterations:]. Le nombre d'itération doit être <= 255.");
	maxIterations = iterations;
	
	[self willChangeValueForKey:@"besoinMAJ"];
	besoinMAJ = YES;
	[self didChangeValueForKey:@"besoinMAJ"];
}

- (int) maxIterations
{
	return maxIterations;	
}


#pragma mark Rendu
/* Renvoie une bitmap de 400x300 pixels, où est représenté l'ensemble de
   Mandelbrot. */
- (NSBitmapImageRep*) bitmapImageRep
{
	// Créer la bitmap
	NSBitmapImageRep* bitmapRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL
																		  pixelsWide:largeurBitmap
																		  pixelsHigh:hauteurBitmap
																	   bitsPerSample:8
																	 samplesPerPixel:1
																			hasAlpha:NO
																			isPlanar:NO
																	  colorSpaceName:NSDeviceWhiteColorSpace
																		 bytesPerRow:0
																		bitsPerPixel:8];
	
	// Obtenir la bitmap
	unsigned char* bitmapPtr = [bitmapRep bitmapData];
	unsigned int rowBytes = [bitmapRep bytesPerRow] - largeurBitmap;
	
	// Déterminer les incréments des coordonnées
	Complexe_t premierPoint, dernierPoint;
	premierPoint = [self complexeAvecCoordBitmapX:0 y:0];
	dernierPoint = [self complexeAvecCoordBitmapX:largeurBitmap-1 y:hauteurBitmap-1];
	
	double incX = (dernierPoint.reel - premierPoint.reel) / largeurBitmap;
	double incY = (dernierPoint.imag - premierPoint.imag) / hauteurBitmap;
	
	// Calculer l'ensemble de Mandelbrot:
	// Parcourir tous les points de la bitmap
	double x, y;
	Complexe_t c = premierPoint;

	for(y = 0; y < hauteurBitmap; y++)
	{		
		for(x = 0; x < largeurBitmap; x++)
		{			
			// Initialiser z[0]
			Complexe_t z = {0.0, 0.0};
			Complexe_t zCarre;
			NSUInteger n;
			for(n=0; n < maxIterations; n++)
			{
				// z[n+1] = z[n+1]^2 + c
					// Mettre z au carré
				zCarre.reel = z.reel*z.reel - z.imag*z.imag;
				zCarre.imag = 2.0 * z.reel * z.imag;  
					// Ajouter c
				z.reel = zCarre.reel + c.reel;
				z.imag = zCarre.imag + c.imag;
				
				// La suite diverge si |z| > 2
				if( (z.reel*z.reel + z.imag*z.imag) > 4.0)
					break;
			}
			
			// Donner le niveau de gris au pixel
			*bitmapPtr = n * 255 / maxIterations;
			bitmapPtr++;
			
			c.reel += incX;
		}
		
		c.reel = premierPoint.reel;
		c.imag += incY;
		bitmapPtr += rowBytes;	// Sauter les octets inutilisés en fin de ligne
	}
	
	[bitmapRep autorelease];
	
	[self willChangeValueForKey:@"besoinMAJ"];
	besoinMAJ = NO;
	[self didChangeValueForKey:@"besoinMAJ"];
	
	return bitmapRep;
}

/* Convertit les coordonnées d'un point sur la bitmap en un nombre complexe (coordonnées
   sur le repère mathématique).
*/
- (Complexe_t) complexeAvecCoordBitmapX:(NSUInteger)bitmapX y:(NSUInteger)bitmapY
{
	Complexe_t point;
	
	double hauteur = (largeur/(double)largeurBitmap) * (double)hauteurBitmap;
	
	point.reel = largeur * ((double)bitmapX/(double)largeurBitmap - 0.5) + centre.reel;
	point.imag = hauteur * (0.5 - (double)bitmapY/(double)hauteurBitmap) + centre.imag; 
	
	return point;
}


#pragma mark Coordonnées
// Multiplie la largeur de l'intervalle calculé par le facteur.
- (void) zoomerDuFacteur:(double)facteur
{
	largeur = largeur * facteur;
	
	// Corriger les valeurs extrêmes
	if(largeur < 0.0001)
		largeur = 0.0001;
	else if(largeur > 8.0)
		largeur = 8.0;
	
	[self willChangeValueForKey:@"besoinMAJ"];
	besoinMAJ = YES;
	[self didChangeValueForKey:@"besoinMAJ"];
}

// Décaler le centre du repère de pixelsHoriz pixels vers la droite et de pixelsVerti pixels vers le haut.
- (void) decalerCentreX:(NSInteger)pixelsHoriz y:(NSInteger)pixelsVerti
{
	// Convertir les décalages en pixels en décalages en coordonnées "repère"
	double facteurReperePixel = largeur / largeurBitmap;
	double deltaX = pixelsHoriz * facteurReperePixel;
	double deltaY = pixelsVerti * facteurReperePixel;
	
	// Décaler le centre du repère
	centre.reel += deltaX;
	centre.imag += deltaY;
	
	[self willChangeValueForKey:@"besoinMAJ"];
	besoinMAJ = YES;
	[self didChangeValueForKey:@"besoinMAJ"];
}

@end
